# -*- coding: utf-8 -*-
import os
import logging
import tempfile

import numpy as np
import pandas as pd

from rastertodataframe import util

log = logging.getLogger(__name__)


def raster_to_dataframe(raster_path, vector_path=None):
    """Convert a raster to a Pandas DataFrame.

    .. note:: This requires enough memory to load the entire raster.

    Parameters
    ----------
    raster_path : str
        Path to raster file.
    vector_path : str
        Optional path to vector file. If given, raster pixels will be extracted
        from features in the vector. If None, all raster pixels are converted
        to a DataFrame.

    Returns
    -------
    pandas.core.frame.DataFrame
    """
    # Placeholders for possible temporary files.
    fid1 = fid2 = tmp_fname = vector_mask_fname = None

    # Get raster band names.
    ras = util.open_raster(raster_path)
    raster_band_names = util.get_raster_band_names(ras)

    # Create a mask from the pixels touched by the vector.
    if vector_path is not None:
        fid1, tmp_fname = tempfile.mkstemp(suffix='.geojson')

        # Add a dummy feature ID column to the vector.
        # This is not always present in OGR features.
        vec_gdf = util.open_vector(vector_path, with_geopandas=True)
        mask_values = list(range(1, len(vec_gdf) + 1))
        vec_gdf['__fid__'] = pd.Series(mask_values)
        vec_gdf.to_file(tmp_fname, driver='GeoJSON')

        # Mask the vector using the feature ID column.
        fid2, vector_mask_fname = tempfile.mkstemp()
        vector_mask = util.burn_vector_mask_into_raster(
            raster_path, tmp_fname, vector_mask_fname,
            vector_field='__fid__')

        # Loop over mask values to extract pixels.
        mask_arr = vector_mask.GetRasterBand(1).ReadAsArray()
        ras_arr = ras.ReadAsArray()  # TODO this is not memory efficient.

        out_dfs = []
        for mask_val in mask_values:
            pixels = util.extract_masked_px(ras_arr, mask_arr, mask_val=mask_val)\
                .transpose()
            fid_px = np.ones(pixels.shape[0]) * mask_val

            mask_df = pd.DataFrame(pixels, columns=raster_band_names)
            mask_df['__fid__'] = fid_px
            out_dfs.append(mask_df)

        # Fill DataFrame with pixels.
        out_df = pd.concat(out_dfs)

        # Join with vector attributes.
        out_df = out_df.merge(vec_gdf, how='left', on='__fid__')

    else:
        # No vector given, simply load the raster.
        ras_arr = ras.ReadAsArray()  # TODO not memory efficient.
        mask_arr = np.ones((ras_arr.shape[1], ras_arr.shape[2]))
        pixels = util.extract_masked_px(ras_arr, mask_arr).transpose()
        out_df = pd.DataFrame(pixels, columns=raster_band_names)

    # TODO screen no data values.

    # Remove temporary files.
    if fid1 is not None and tmp_fname is not None:
        os.close(fid1)
        os.remove(tmp_fname)
    if fid2 is not None and vector_mask_fname is not None:
        os.close(fid2)
        os.remove(vector_mask_fname)

    # Return dropping any extra cols.
    return out_df.drop(columns=['__fid__', 'geometry'], errors='ignore')



# -*- coding: utf-8 -*-
import logging

import pyproj
import geopandas as gpd
from osgeo import gdal, ogr, osr

log = logging.getLogger(__name__)


def open_raster(path, read_only=True):
    """Open a raster using GDAL.

    Parameters
    ----------
    path : str
        Path of file to open.
    read_only : bool
        File mode, set to False to open in "update" mode.

    Returns
    -------
    GDAL dataset
    """
    access = gdal.GA_ReadOnly if read_only else gdal.GA_Update
    return gdal.OpenShared(path, access)


def open_vector(path, with_geopandas=False, read_only=True):
    """Open a vector dataset using OGR or GeoPandas.

    Parameters
    ----------
    path : str
        Path to vector file.
    with_geopandas : bool
        Set to True to open with geopandas, else use OGR.
    read_only : bool
        If opening with OGR, set to False to open in "update" mode.

    Returns
    -------
    GeoDataFrame if ``with_geopandas`` else OGR datsource.
    """
    if with_geopandas:
        return gpd.read_file(path)

    update = False if read_only else True
    return ogr.OpenShared(path, update=update)


def _get_dataset_epsg(dataset):
    """Get the EPSG code from a GDAL Dataset.

    Parameters
    ----------
    dataset : gdal.Dataset

    Returns
    -------
    int
    """
    prj = dataset.GetProjection()
    return _epsg_from_projection(prj)


def _get_datasource_epsg(datasource):
    """Get the EPSG code from a OGR DataSource.

    Parameters
    ----------
    datasource : ogr.DataSource

    Returns
    -------
    int
    """
    layer = datasource.GetLayer(0)
    spatial_ref = layer.GetSpatialRef()
    spatial_ref.AutoIdentifyEPSG()
    return int(spatial_ref.GetAuthorityCode(None))


def _get_gpd_epsg(gdf):
    """Get the EPSG code from a GeoPandas DataFrame.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame

    Returns
    -------
    int
    """
    return _epsg_from_projection(pyproj.Proj(gdf.crs).srs)


def _epsg_from_projection(prj):
    """Return the EPSG code from a projection string.

    Parametres
    ----------
    prj : str

    Returns
    -------
    int
    """
    srs = osr.SpatialReference()

    if prj.startswith("PROJCS") or prj.startswith("GEOGCS"):
        # ESRI well know text strings,
        srs.ImportFromESRI([prj])
    else:
        # Proj4 strings.
        srs.ImportFromProj4(prj)

    srs.AutoIdentifyEPSG()
    return int(srs.GetAuthorityCode(None))


def get_epsg(data):
    """Get the EPSG code from an input.

    Parameters
    ----------
    data : gdal.Dataset or ogr.DataSource or gpd.GeoDataFrame

    Returns
    -------
    int
    """
    if isinstance(data, gdal.Dataset):
        return _get_dataset_epsg(data)
    elif isinstance(data, ogr.DataSource):
        return _get_datasource_epsg(data)
    elif isinstance(data, gpd.GeoDataFrame):
        return _get_gpd_epsg(data)
    else:
        raise ValueError('Unable to get EPSG from: {}'.format(data))


def same_epsg(data1, data2):
    """Check sets of data have the same EPSG.

    Parameters
    ----------
    data1 : gdal.DataSet or ogr.DataSource or gpd.GeoDataFrame
    data2 : gdal.DataSet or ogr.DataSource or gpd.GeoDataFrame

    Returns
    -------
    bool
    """
    return get_epsg(data1) == get_epsg(data2)


def _create_empty_raster(template, out_path, n_bands=1, no_data_value=None):
    """Create a new empty raster using GDAL. Inherits all but the data from the
        given template dataset.

    Parameters
    ----------
    template : gdal.Dataset
    out_path : str
    n_bands : int or None
        Number of bands to create in the output raster.
    no_data_value : float or None
        No data value, if None uses the same as ``template``.

    Returns
    -------
    gdal.DataSet
    """
    # Raster size.
    x_size = template.RasterXSize
    y_size = template.RasterYSize
    n_bands = int(n_bands) if n_bands is not None else template.RasterCount
    dtype = template.GetRasterBand(1).DataType

    # Create the driver.
    driver = template.GetDriver()
    out_dataset = driver.Create(out_path, x_size, y_size, n_bands, dtype)

    # Set the projection.
    out_dataset.SetGeoTransform(template.GetGeoTransform())
    out_dataset.SetProjection(template.GetProjectionRef())

    # Set the no data value.
    if no_data_value is not None:
        out_ndv = float(no_data_value)
        for i in range(1, out_dataset.RasterCount + 1):
            band = out_dataset.GetRasterBand(i)
            band.SetNoDataValue(out_ndv)

    return out_dataset


def burn_vector_mask_into_raster(raster_path, vector_path, out_path,
                                 vector_field=None):
    """Create a new raster based on the input raster with vector features
        burned into the raster. To be used as a mask for pixels in the vector.

    Parameters
    ----------
    raster_path : str
    vector_path : str
    out_path : str
        Path for output raster. Format and Datatype are the same as ``ras``.
    vector_field : str or None
        Name of a field in the vector to burn values from. If None, all vector
        features are burned with a constant value of 1.

    Returns
    -------
    gdal.Dataset
        Single band raster with vector geometries burned.
    """

    ras = open_raster(raster_path)
    vec = open_vector(vector_path)

    # Check EPSG are same, if not reproject vector.
    if not same_epsg(ras, vec):
        raise ValueError(
            'Raster and vector are not the same EPSG.\n'
            '{} != {}'.format(get_epsg(ras), get_epsg(vec))
        )

    # Create an empty for GDALRasterize to burn vector values to.
    out_ds = _create_empty_raster(ras, out_path, n_bands=1, no_data_value=0)

    if vector_field is None:
        # Use a constant value for all features.
        burn_values = [1]
        attribute = None
    else:
        # Use the values given in the vector field.
        burn_values = None
        attribute = vector_field

    # Options for Rasterize.
    # note: burnValues and attribute are exclusive.
    rasterize_opts = gdal.RasterizeOptions(
        bands=[1],
        burnValues=burn_values,
        attribute=attribute,
        allTouched=True)
    _ = gdal.Rasterize(out_ds, vector_path, options=rasterize_opts)

    # Explicitly close raster to ensure it is saved.
    out_ds.FlushCache()
    out_ds = None

    return open_raster(out_path)


def get_raster_band_names(raster):
    """Obtain the names of bands from a raster. The raster metadata is queried
    first, if no names a present, a 1-index list of band_N is returned.

    Parameters
    ----------
    raster : gdal.Dataset

    Returns
    -------
    list[str]
    """
    band_names = []
    for i in range(1, raster.RasterCount + 1):
        band = raster.GetRasterBand(i)

        if band.GetDescription():
            # Use the band description.
            band_names.append(band.GetDescription())
        else:
            # Check for metedata.
            this_band_name = 'Band_{}'.format(band.GetBand())
            metadata = band.GetDataset().GetMetadata_Dict()

            # If in metadata, return the metadata entry, else Band_N.
            if this_band_name in metadata and metadata[this_band_name]:
                band_names.append(metadata[this_band_name])
            else:
                band_names.append(this_band_name)

    return band_names


def extract_masked_px(ras, mask, mask_val=None):
    """Extract raster values according to a mask.

    Parameters
    ----------
    ras : np.ndarray
        2D or 3D input data in the form [bands][y][x].
    mask : np.ndarray
        1D or 2D with zeroes to mask data.
    mask_val : int
        Value of the data pixels in the mask. Default: non-zero.

    Returns
    -------
    np.ndarray
        Array of non-masked data.
    """
    if mask is None:
        return ras

    # Use the mask to get the indices of the non-zero pixels.
    if mask_val:
        (i, j) = (mask == mask_val).nonzero()
    else:
        (i, j) = mask.nonzero()

    return (ras[i, j] if ras.ndim == 2 else ras[:, i, j])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `rastertodataframe.util` package."""

import os
import unittest
import tempfile

from osgeo import gdal, ogr
import geopandas as gpd

from rastertodataframe import util


class TestRasterToDataFrameUtil(unittest.TestCase):
    def setUp(self):
        test_data_path = os.path.join(os.path.dirname(__file__))
        self.raster_path = os.path.join(test_data_path, 'raster.tif')
        self.vector_path = os.path.join(test_data_path, 'vector.geojson')
        self.raster_wgs84_path = os.path.join(test_data_path,
                                              'raster_epsg4326.tif')

    def test_open_raster(self):
        ras = util.open_raster(self.raster_path)
        self.assertIsInstance(ras, gdal.Dataset)

    def test_open_vector_ogr(self):
        # As OGR DataSource.
        vec = util.open_vector(
            self.vector_path,
            with_geopandas=False)
        self.assertIsInstance(vec, ogr.DataSource)

    def test_open_vector_gdf(self):
        # As GeoPandasDataFrame.
        vec_gdf = util.open_vector(
            self.vector_path,
            with_geopandas=True)
        self.assertIsInstance(vec_gdf, gpd.GeoDataFrame)

    def test__epsg_from_projection_proj4(self):
        # Proj4 string.
        prj = '+proj=utm +zone=32 +datum=WGS84 +units=m +no_defs'
        self.assertEqual(
            util._epsg_from_projection(prj),
            32632)

    def test__epsg_from_projection_esri(self):
        # ESRI PROJCS (WKT).
        prj = ('PROJCS["WGS_1984_UTM_Zone_30N", '
               'GEOGCS["GCS_WGS_1984", '
               'DATUM["D_WGS_1984", '
               'SPHEROID["WGS_1984", 6378137.0, 298.257223563]], '
               'PRIMEM["Greenwich", 0.0], '
               'UNIT["Degree", 0.0174532925199433]], '
               'PROJECTION["Transverse_Mercator"], '
               'PARAMETER["False_Easting", 500000.0], '
               'PARAMETER["False_Northing", 0.0], '
               'PARAMETER["Central_Meridian", -3.0], '
               'PARAMETER["Scale_Factor", 0.9996], '
               'PARAMETER["Latitude_Of_Origin", 0.0], '
               'UNIT["Meter", 1.0]]')

        self.assertEqual(
            util._epsg_from_projection(prj),
            32630)

    def test_get_ogr_epsg(self):
        vec = ogr.OpenShared(self.vector_path)
        self.assertEqual(util.get_epsg(vec), 4326)

    def test_get_gdal_epsg(self):
        ras = gdal.OpenShared(self.raster_path)
        self.assertEqual(util.get_epsg(ras), 32632)

    def test_get_gpd_epsg(self):
        gdf = gpd.read_file(self.vector_path)
        self.assertEqual(util.get_epsg(gdf), 4326)

    def test_same_epsg(self):
        ras = gdal.OpenShared(self.raster_path)
        gdf = gpd.read_file(self.vector_path)
        self.assertFalse(util.same_epsg(ras, gdf))
        self.assertTrue(util.same_epsg(ras, ras))

    def test__create_empty_raster(self):
        fid, tmp_fname = tempfile.mkstemp()
        ras = gdal.OpenShared(self.raster_path)
        out = util._create_empty_raster(
            ras, tmp_fname, n_bands=1, no_data_value=0)

        self.assertIsInstance(out, gdal.Dataset)
        self.assertEqual(out.RasterCount, 1)
        self.assertEqual(ras.RasterXSize, out.RasterXSize)
        self.assertEqual(ras.RasterYSize, out.RasterYSize)

        os.close(fid)
        os.remove(tmp_fname)

    def test__burn_vector_mask_into_raster_wrong_epsg(self):
        # Error for differing projections.
        with self.assertRaises(ValueError):
            out = util.burn_vector_mask_into_raster(
                self.raster_path, self.vector_path, '')

    def test__burn_vector_mask_into_raster_vector_mask(self):
        # Burn in a specified vector field.
        fid, tmp_fname = tempfile.mkstemp()
        out = util.burn_vector_mask_into_raster(
            self.raster_wgs84_path, self.vector_path, tmp_fname)

        arr = out.GetRasterBand(1).ReadAsArray()
        self.assertEqual(arr.shape, (39, 58))
        self.assertEqual(arr.ndim, 2)
        self.assertEqual(arr.max(), 1)

        os.close(fid)
        os.remove(tmp_fname)

    def test__burn_vector_mask_into_raster_vector_field(self):
        # Burn in a specified vector field.
        fid, tmp_fname = tempfile.mkstemp()
        out = util.burn_vector_mask_into_raster(
            self.raster_wgs84_path, self.vector_path, tmp_fname,
            vector_field='value')

        arr = out.GetRasterBand(1).ReadAsArray()
        self.assertEqual(arr.shape, (39, 58))
        self.assertEqual(arr.ndim, 2)
        self.assertEqual(arr.max(), 2000)

        os.close(fid)
        os.remove(tmp_fname)

    def test_get_raster_band_names(self):
        ras = gdal.OpenShared(self.raster_path)
        band_names = util.get_raster_band_names(ras)
        expected = ['Band_1', 'Band_2', 'Band_3', 'Band_4']
        self.assertListEqual(band_names, expected)

    def test_extract_mask_px(self):
        # TODO
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `rastertodataframe` package."""

import os
import unittest
import tempfile

from osgeo import gdal, ogr
import geopandas as gpd

from rastertodataframe import raster_to_dataframe


class TestRasterToDataFrame(unittest.TestCase):
    def setUp(self):
        test_data_path = os.path.join(os.path.dirname(__file__))
        self.raster_path = os.path.join(test_data_path, 'raster.tif')
        self.vector_path = os.path.join(test_data_path, 'vector.geojson')
        self.raster_wgs84_path = os.path.join(test_data_path,
                                              'raster_epsg4326.tif')

    def test_raster_to_dataframe_with_vector(self):
        out_df = raster_to_dataframe(
            self.raster_wgs84_path, vector_path=self.vector_path)

        expected_cols = ['Band_1', 'Band_2', 'Band_3', 'Band_4',
                         'fid', 'value', 'value_string']

        self.assertEqual(out_df.shape, (267, 7))
        self.assertCountEqual(list(out_df.columns), expected_cols)

    def test_raster_to_dataframe_without_vector(self):
        out_df = raster_to_dataframe(self.raster_path)

        expected_cols = ['Band_1', 'Band_2', 'Band_3', 'Band_4']

        self.assertEqual(out_df.shape, (2204, 4))
        self.assertCountEqual(list(out_df.columns), expected_cols)



# -*- coding: utf-8 -*-

"""Top-level package for Raster To DataFrame."""

__author__ = """Martin Black"""
__email__ = 'mblack@posteo.de'
__version__ = '0.1.1'

from .rastertodataframe import *
from . import util
